/* ///////////////////////////////////////////////////////////////////// 
                              Orbit  (v. 0.6)
            
    Solve the equations of motion for a charged particle in  
    a (user-supplied) electro-magnetic field:
 
      dx/dt = v
      dv/dt = q/m*(E + v/c X B)

    where q and m are the particle's charge and mass, c is the speed
    of light.
    The actual equations are integrated in dimensionless form by 
    setting v' = v/c, t' = Omega*t, x' = x*Omega/c where Omega = q*B0/(m*c) 
    is the Larmor frequency corresponding to a reference magnetic field
    unit B0 (example B0 = 1 Gauss).

     dx'/dt' = v'
     dv'/dt' = (E + v' X B)    
   
    Author:         A. Mignone (andrea.mignone@unito.it)
    Last Modified:  Oct 10, 2022
   ///////////////////////////////////////////////////////////////////// */
#include "orbit.h"

/* ------------------------------------------------------------------
    Use testNumber to select the EM field configuration.
      1: Uniform B
      2: EXB Drift
      3: Curvature Drift
      4: Gradient Drift
      5: Magnetic Mirror
      6: Magnetic Dipole
   ------------------------------------------------------------------ */

static int testNumber = 1;

int main()
{
  int    nv, err;
  long int  n = 0, nok = 0;
  double t, dt, tstop, dt_next, tol = 1.e-9;
  double dt_max = 0.05;
  double q[NVAR], maxq[NVAR], minq[NVAR];
  char input_string[256], test_name[64];
  FILE *fin, *fout;

/* ------------------------------------------------------------
   0. Print some info
   ------------------------------------------------------------ */

  printf ("Orbit Code (v 0.5) - by A. Mignone\n");

/* ------------------------------------------------------------
   1. Read input parameter file 'orbit.ini'
   ------------------------------------------------------------ */

  printf ("Reading input file orbit.ini...\n");
  fin = fopen("orbit.ini","r");
  if (fin == NULL){
    printf ("! Error: 'orbit.dat' not found\n");
    exit(1);
  }
  while (!feof(fin)){
    fscanf (fin, "%s",input_string);
    if (!strcmp(input_string,"Case:")){
      fscanf(fin, "%d\n", &nv);
      if (nv == testNumber) break;
    }
  }

  fgets (test_name, 32, fin);
  test_name[strlen(test_name)-2] = '\0'; // Redefine string by removing last char

  fscanf (fin, "%s  %lf  %lf  %lf\n", input_string, q+XX, q+YY, q+ZZ);
  fscanf (fin, "%s  %lf  %lf  %lf\n", input_string, q+VX, q+VY, q+VZ);
  fscanf (fin, "%s  %lf\n", input_string, &tstop);
  fclose(fin);
  VAR_LOOP(nv) maxq[nv] = minq[nv] = q[nv];

  printf ("[x(0),   y(0),  z(0)] = %8.3e  %8.3e  %8.3e\n",q[XX], q[YY], q[ZZ]);
  printf ("[vx(0), vy(0), vz(0)] = %8.3e  %8.3e  %8.3e\n",q[VX], q[VY], q[VZ]);
  printf ("tstop = %8.3e\n\n",tstop);

/* ------------------------------------------------------------
   2. Start integration loop 
   ------------------------------------------------------------ */

  printf ("Starting computation\n");
  fout = fopen("orbit.dat","w");
  t  = 0.0; 
  dt = 1.e-4;
  fprintf (fout,"#    t \t\t x\t\t y\t\t z\n");
  fprintf (fout, "%12.6e  %12.6e  %12.6e  %12.6e\n",t, q[XX], q[YY], q[ZZ]);
  while (t < tstop){
    
    err = CK45(q, t, dt, &dt_next, tol);
    if (err == 0){  /* -- integration successful: keep going -- */
      t  += dt;
      fprintf (fout,"%12.6e  %12.6e  %12.6e  %12.6e\n",t, q[XX], q[YY], q[ZZ]);
      VAR_LOOP(nv) maxq[nv] = MAX(maxq[nv], q[nv]);
      VAR_LOOP(nv) minq[nv] = MIN(minq[nv], q[nv]);
      nok++;
    }
    n++;
    if (n == NMAX_STEPS) break;
    dt = MIN(dt_next, dt_max);
  }
  fclose(fout);

/* -----------------------------------------------------------
   3. Print some integration information
   ----------------------------------------------------------- */

  printf ("Total number of steps: %ld\n",n);
  printf ("Total number of (successful) steps: %ld\n",nok);
  printf ("Xrange: [%f, %f]\n",minq[XX],maxq[XX]);
  printf ("Yrange: [%f, %f]\n",minq[YY],maxq[YY]);
  printf ("Zrange: [%f, %f]\n",minq[ZZ],maxq[ZZ]);

  printf ("[x]   : %12.6e (cm)\n",c_CONST/Omega1_CONST);
  printf ("[t]   : %12.6e (s)\n",1.0/Omega1_CONST);
  
  printf ("Data written to 'orbit.dat'\n");
  printf ("Done.\n");

/* -----------------------------------------------------------
   4. Set plot environments for gnuplot users
   ----------------------------------------------------------- */
 
  fout = fopen("plot_setting.gp","w");
  fprintf (fout,"nsteps = %ld\n",nok);
  fprintf (fout,"xmin = %12.6e; xmax = %12.6e\n",minq[XX],maxq[XX]);
  fprintf (fout,"ymin = %12.6e; ymax = %12.6e\n",minq[YY],maxq[YY]);
  fprintf (fout,"zmin = %12.6e; zmax = %12.6e\n",minq[ZZ],maxq[ZZ]);
  fprintf (fout,"title = '%s'\n",test_name);  
  fclose(fout);

  return 0;
}

/* ********************************************************************* */
void EMFields (double *E, double *B,
               double x, double y, double z, double t)
/*
 * Assign the electric and magnetic fields.
 *
 *********************************************************************** */
{

/* -- constant uniform B -- */

  if (testNumber == 1){
    E[XX] = 0.0;
    E[YY] = 0.0;
    E[ZZ] = 0.0;

    B[XX] = 0.0;
    B[YY] = 0.0;
    B[ZZ] = 1.0;
    return;
  }

/* -- B, E constant, E perpendicular to B --  */

  if (testNumber == 2){ 
    E[XX] = 0.0;
    E[YY] = 0.01;
    E[ZZ] = 0.0;

    B[XX] = 0.0;
    B[YY] = 0.0;
    B[ZZ] = 1.0;
    return;
  }

/* -- curvature drift -- */

  if (testNumber == 3){
    double z0 = 0.0, y0 = -2.0, r;
    r = sqrt((y-y0)*(y-y0) + (z-z0)*(z-z0));
    E[XX] = 0.0;
    E[YY] = 0.0;
    E[ZZ] = 0.0;

    B[XX] = 0.0;
    B[YY] = -(z-z0)/r;
    B[ZZ] =  (y-y0)/r;
    return;
  }

/* -- gradient drift -- */

  if (testNumber == 4){
    E[XX] = 0.0;
    E[YY] = 0.0;
    E[ZZ] = 0.0;

    B[XX] = 0.0;
    B[YY] = 0.0;
    B[ZZ] = 1.0 + 2*y;
    return;
  }

/* -- magnetic mirror -- */

  if (testNumber == 5){
    double r = sqrt(x*x + y*y)+1.e-8, a = 1.0;
    double cphi = x/r;
    double sphi = y/r;
    double Br, Bz, Bm = 50;
/*
    Br = z*exp(-(z*z)/(a*a))*log(1.0 + 0*r*r)/(r*a*a)*(Bm-1.0);
    Bz = 1.0 + (Bm-1.0)*exp(-(z*z)/(a*a))/(1.0 + r*r);
*/
    Br = r*z*exp(-(z*z)/(a*a))*(Bm-1.0);
    Bz = 1.0 + (Bm-1.0)*exp(-(z*z)/(a*a));

    E[XX] = 0.0;
    E[YY] = 0.0;
    E[ZZ] = 0.0;

    B[XX] =  Br*cphi;
    B[YY] =  Br*sphi;
    B[ZZ] =  Bz;
    return;
  }

/* -- magnetic dipole -- */

  if (testNumber == 6){
    double r  = sqrt(x*x + y*y + z*z)+1.e-12;
    double r2 = r*r;
    double r5 = r2*r2*r;
    double M  = 5.0;

    E[XX] = 0.0;
    E[YY] = 0.0;
    E[ZZ] = 0.0;

    B[XX] = 3.0*M*x*z/r5;
    B[YY] = 3.0*M*y*z/r5;
    B[ZZ] = M*(3.0*z*z - r*r)/r5;
    return;
  }

  printf ("! Unknown test name case\n");
  exit(1);
/*
restart:
Bz := 1 + (Bm-1)*exp(-(z/a)^2)/(1+0*r^2);
dBz_dz := diff(Bz,z):
Br := -1/r*int(r*dBz_dz,r=0..r);
limit(Br,r=0);
dB := simplify(dBz_dz + 1/r*diff(r*Br,r));
*/
}

/* ********************************************************************* */
void RHS (double *q, double t, double *rhs)
/*
 * Compute the right hand side of the equations.
 *
 *********************************************************************** */
{
  double x, y, z, E[3], B[3];

  x = q[XX];
  y = q[YY];
  z = q[ZZ];
  EMFields (E, B, x, y, z, t);

  rhs[XX] = q[VX];
  rhs[YY] = q[VY];
  rhs[ZZ] = q[VZ];

  rhs[VX] = E[XX] + q[VY]*B[ZZ] - q[VZ]*B[YY];
  rhs[VY] = E[YY] + q[VZ]*B[XX] - q[VX]*B[ZZ];
  rhs[VZ] = E[ZZ] + q[VX]*B[YY] - q[VY]*B[XX];
}

/* ********************************************************************* */
int CK45 (double *q0, double t, double dt, double *dt_next, double tol)
/*!
 *
 *  Solve the system of equations using an explicit Cash-Karp 4-5
 *  integrator.
 *
 * The return status can be one of
 *
 *  0              = success
 *  ERR_StepFailed = step has large error. 
 *  ERR_RHS        = an error occured in computing the right hand side.
 * 
 *********************************************************************** */
{
  int   i, n, rhs_err;

  double scrh, err, dt_shrink, dt_grow;
  double qscal;

  const double a2 = 0.2;
  const double a3 = 0.3;
  const double a4 = 0.6;
  const double a5 = 1.0;
  const double a6 = 0.875;

  const double c1 = 37.0/378.0;
  const double c2 = 0.0;
  const double c3 = 250.0/621.0;
  const double c4 = 125.0/594.0;
  const double c5 = 0.0;
  const double c6 = 512.0/1771.0; 

  const double cs1 = 2825.0/27648.0;
  const double cs2 = 0.0;
  const double cs3 = 18575.0/48384.0;
  const double cs4 = 13525.0/55296.0;
  const double cs5 = 277.0/14336.0;
  const double cs6 = 0.25;

  const double b21 = 0.2;
  const double b31 = 3.0/40.0      , b32 = 9.0/40.0;
  const double b41 = 0.3           , b42 = -0.9       , b43 = 1.2;
  const double b51 = -11.0/54.0    , b52 = 2.5        , b53 = -70.0/27.0   , b54 = 35.0/27.0;
  const double b61 = 1631.0/55296.0, b62 = 175.0/512.0, b63 = 575.0/13824.0, b64 = 44275.0/110592.0, b65 = 253.0/4096.0;
    
  double q1[NVAR], qs[NVAR], q4th[NVAR], q5th[NVAR];
  double k1[NVAR], k2[NVAR], k3[NVAR]; 
  double k4[NVAR], k5[NVAR], k6[NVAR];

/* -------------------------------------------------------
    Make a copy of the solution vector.
   ------------------------------------------------------- */

  VAR_LOOP(n) q1[n] = q0[n];

/* -------------------------------------------------------
                Step cycle
   -------------------------------------------------------- */

  RHS (q1, t, k1);
  VAR_LOOP(n) qs[n] = q1[n] + dt*b21*k1[n];

  RHS (qs, t + a2*dt, k2);
  VAR_LOOP(n) qs[n] = q1[n] + dt*(b31*k1[n] + b32*k2[n]);

  RHS (qs, t + a3*dt, k3);
  VAR_LOOP(n) qs[n] = q1[n] + dt*(b41*k1[n] + b42*k2[n] + b43*k3[n]);

  RHS (qs, t + a4*dt, k4);
  VAR_LOOP(n) qs[n] = q1[n] + dt*(b51*k1[n] + b52*k2[n] + b53*k3[n] 
                                            + b54*k4[n]);
  RHS (qs, t + a5*dt, k5);
  VAR_LOOP(n) qs[n] = q1[n] + dt*(b61*k1[n] + b62*k2[n] + b63*k3[n] 
                                            + b64*k4[n] + b65*k5[n]);

  RHS (qs, t + a6*dt, k6);

/* -- compute 5th and 4th order solutions -- */

  VAR_LOOP(n) {
    q5th[n] = q1[n] + dt*(c1*k1[n] + c2*k2[n] + c3*k3[n] 
                        + c4*k4[n] + c5*k5[n] + c6*k6[n]);
    q4th[n] = q1[n] + dt*(cs1*k1[n] + cs2*k2[n] + cs3*k3[n] 
                        + cs4*k4[n] + cs5*k5[n] + cs6*k6[n]);
  }

/* --------------------------------------------------------
    compute error

    the following should give an absolute error if the
    solution is close to zero and a relative one in the 
    limit of large |q|
   -------------------------------------------------------- */

  err   = 0.0;
  VAR_LOOP(n) {
    scrh = fabs(q5th[n] - q4th[n])/(1.0 + fabs(q1[n]));
    err  = MAX(err, scrh);
/*
    scrh = creal(q5th[n] - q4th[n])/(1.0 + creal(q0[n]));
    err  = max(err, scrh);

    scrh = cimag(q5th[n] - q4th[n])/(1.0 + cimag(q0[n]));
    err  = max(err, scrh);
*/
  }
  err /= tol;

  if (err < 1.0){  /* -- ok, accept step -- */

    err = MAX(err, 1.e-18);

  /* -- provide an estimate for next dr -- */

    dt_grow  = 0.9*fabs(dt)*pow(err, -0.2);
    dt_grow  = MIN(dt_grow, 5.0*fabs(dt)); /* -- do not increase more than 5 -- */
    *dt_next = dt_grow;

    VAR_LOOP(n) q0[n] = q1[n] = q5th[n];
    return 0;

  }else{   /* -- shrink dr and redo time step -- */

    dt_shrink = 0.9*fabs(dt)*pow(err, -0.25);
    dt        = MAX(dt_shrink, 0.05*fabs(dt)); /* -- do not decrease more than 20 -- */
    *dt_next  = dt;
    return -1;  /* step failed */
  } 
}

void Show(double *q)
{
  printf ("> %12.6e  %12.6e  %12.6e  %12.6e  %12.6e  %12.6e\n",
          q[XX], q[YY], q[ZZ], q[VX], q[VY], q[VZ]);

}

